"""
B{Course:}     TKP4550 Master's Thesis
B{Task:}       Computes an automaton w.r.t a stateindex given intersection points and stateindex

@author      : Mandar Thombre
@contact     : mandart@stud.ntnu.no
@organization: Department of Chemical Engineering, NTNU, Norway
@since       : Nov 1, 2016
@change      : June 15, 2017
@version     : 1.2
@todo        : see comments at the bottom
@requires    : Python v2.7
"""


import numpy as np
import copy

## Assumption! For now, I assume I have all the points of intersection between hypercubes and equilibrium 
## hypersurfaces, and they are not on intersection of gridlines

class Grid:
	
	def __init__(self, no_of_states, boundaries):
		self.no_of_states = no_of_states
		self.boundaries = boundaries
		self.partitions = [len(_)-1 for _ in boundaries]

	def ComputeGridSquare(self, x):
		'''
		Computes the grid square(hypercube) that a given point x lies in
		Indexing of grid in this representation is {(1..n1) X (1..n2) X (1..n3) X ....}
		'''
		gridindex = [0]*self.no_of_states
		
		for stateindex in range(self.no_of_states):
			if min(self.boundaries[stateindex]) < x[stateindex] < max(self.boundaries[stateindex]):
				gridindex[stateindex] = np.searchsorted(self.boundaries[stateindex],x[stateindex])
			else:
				raise ValueError("Point outside grid!")
		
		return gridindex


	def NextGridSquare(self, gridindex, stateindex):
		'''
		Takes to next grid square (+ve) in the given stateindex
		'''

		newgridindex = copy.deepcopy(gridindex)
		if gridindex[stateindex] < self.partitions[stateindex]:
			newgridindex[stateindex] = newgridindex[stateindex]+1
		else:
			raise ValueError("Can't go forward! Grid in this state direction ends here!")

		return newgridindex

	def PrevGridSquare(self, gridindex, stateindex):
		'''
		Takes to previous grid square (-ve) in the given stateindex
		'''
		
		newgridindex = copy.deepcopy(gridindex)
		if gridindex[stateindex] > 1:
			newgridindex[stateindex] = newgridindex[stateindex]-1
		else:
			raise ValueError("Can't go backword! Grid in this state direction ends here!")

		return newgridindex
	
	def PullApart(self, x):
		'''
		Pulls apart the given point x in each state direction
		'''
		epsilon = 1e-12 #some very small number
		
		xplus =  [0]*self.no_of_states
		xminus = [0]*self.no_of_states
		
		for stateindex in range(self.no_of_states):
			if x[stateindex] == min(self.boundaries[stateindex]):
				xplus[stateindex]  = x[stateindex]+epsilon
				xminus[stateindex] = x[stateindex]+epsilon
			if x[stateindex] == max(self.boundaries[stateindex]):
				xplus[stateindex]  = x[stateindex]-epsilon
				xminus[stateindex] = x[stateindex]-epsilon
			if min(self.boundaries[stateindex]) < x[stateindex] < max(self.boundaries[stateindex]):
				xplus[stateindex]  = x[stateindex]+epsilon
				xminus[stateindex] = x[stateindex]-epsilon

		return xplus, xminus


def cartesian(arrays, out=None):
    """
    Generate a cartesian product of input arrays.
    """

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = n / arrays[0].size
    out[:,0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m,1:])
        for j in xrange(1, arrays[0].size):
            out[j*m:(j+1)*m,1:] = out[0:m,1:]
    return out

if __name__ == '__main__': 

	## 3D system
	boundaries1 = [0,1,2,3]
	boundaries2 = [0,1,2,3,4]
	boundaries3 = [0,1,2]

	g = Grid(3,[boundaries1,boundaries2, boundaries3])

	automaton    = np.zeros(g.partitions)
	type3squares = []

	## intersection points
	intpoints    = [[1.5,4,0], [1.5,3,0], [1,2.5,0], [1.5,4,1], [1.5,3,1], [1,2.5,1], [1.5,4,2], [1.5,3,2], [1,2.5,2]]
	pulledpoints = []

	for point in intpoints:
		xplus, xminus = g.PullApart(point)
		pulledpoints.append(xplus)
		pulledpoints.append(xminus)

	for point in pulledpoints:
		gridindex = g.ComputeGridSquare(point)
		if gridindex not in type3squares:
			type3squares.append(gridindex)
		
		grid_python_index  = [i-1 for i in gridindex] #for indexing in python, 0 start
		automaton[tuple(grid_python_index)] = 3

	## Automaton based on stateindex = 0
	## This part won't fill up any array in the direction of stateindex that does not have a type3square 
	automaton_stateindex = 0
	for gsq in type3squares:	
		tempgsq = copy.deepcopy(gsq)
		while tempgsq[automaton_stateindex] < g.partitions[automaton_stateindex]:
			tempgsq = g.NextGridSquare(tempgsq, automaton_stateindex)
			nextgsq_python_index  = [i-1 for i in tempgsq]
			if automaton[tuple(nextgsq_python_index)] == 0:
				automaton[tuple(nextgsq_python_index)] = 2

		tempgsq = copy.deepcopy(gsq)
		while tempgsq[automaton_stateindex] > 1:
			tempgsq = g.PrevGridSquare(tempgsq, automaton_stateindex)
			prevgsq_python_index  = [i-1 for i in tempgsq] 
			if automaton[tuple(prevgsq_python_index)] == 0:
				automaton[tuple(prevgsq_python_index)] = 1


	## Now we fill up all the empty arrays
	allsquares   = cartesian([range(1,g.partitions[i]+1) for i in range(g.no_of_states)])
	zerosquares  = []

	for gsq in allsquares:
		gsq_python_index = [i-1 for i in gsq]
		if automaton[tuple(gsq_python_index)] == 0:
			zerosquares.append(gsq)

	iflag = 0
	motherload = None
	for gsq in zerosquares:
		for stateindex in range(g.no_of_states):
			
			tempgsq = copy.deepcopy(gsq)
			while tempgsq[stateindex] < g.partitions[stateindex]:
				tempgsq = g.NextGridSquare(tempgsq, stateindex)
				nextgsq_python_index  = [i-1 for i in tempgsq] # for python
				
				if automaton[tuple(nextgsq_python_index)] == 3:
					break

				if automaton[tuple(nextgsq_python_index)] != 0 and automaton[tuple(nextgsq_python_index)] != 3:
					fillall = automaton[tuple(nextgsq_python_index)]
					iflag = 1
					break
			
			if iflag == 1:
				break
		
		if iflag == 1:
			break

			tempgsq = copy.deepcopy(gsq)
			while tempgsq[stateindex] > 1:
				tempgsq = g.PrevGridSquare(tempgsq, stateindex)
				prevgsq_python_index  = [i-1 for i in tempgsq] # for python

				if automaton[tuple(nextgsq_python_index)] == 3:
					break

				if automaton[tuple(prevgsq_python_index)] != 0 and automaton[tuple(prevgsq_python_index)] != 3:
					fillall = automaton[tuple(prevgsq_python_index)]
					iflag = 1
					break
			if iflag == 1:
				break
		if iflag == 1:
			break

	for gsq in zerosquares:
		gsq_python_index = [i-1 for i in gsq]
		automaton[tuple(gsq_python_index)] = fillall

	print automaton

	'''
	The presence of "nan" in the automaton indicates that the equilibrium hypersurface is "parallel" 
	or "close to parallel"	to direction of chosen stateindex, since automaton computation is based 
	only on intersection points	(and corresponding grid hypercubes) and not the surface itself.	

	This happens when, before the fillall calculation, arrays along a stateindex are either all 3 or all 0.
	Knowledge of the hypersurface may be exploited here. For example, the normal from any point in any of the 
	'zerosquares' to the hypersurface may be calculated and the sign of normal would determine the state directionality
	of all zerosquares (1 or 2). This is straightforward for linear hypersurfaces but trickier for non linear hypersurfaces
	(future implementation)
	'''