# Copyright belongs to David Eppstein. The code has been modified for use according to the examples in this master's thesis.


# Dijkstra's algorithm for shortest paths
# David Eppstein, UC Irvine, 4 April 2002

# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/117228
from __future__ import generators

class priorityDictionary(dict):
    def __init__(self):
        '''Initialize priorityDictionary by creating binary heap
of pairs (value,key).  Note that changing or removing a dict entry will
not remove the old pair from the heap until it is found by smallest() or
until the heap is rebuilt.'''
        self.__heap = []
        dict.__init__(self)

    def smallest(self):
        '''Find smallest item after removing deleted items from heap.'''
        if len(self) == 0:
            raise IndexError, "smallest of empty priorityDictionary"
        heap = self.__heap
        while heap[0][1] not in self or self[heap[0][1]] != heap[0][0]:
            lastItem = heap.pop()
            insertionPoint = 0
            while 1:
                smallChild = 2*insertionPoint+1
                if smallChild+1 < len(heap) and \
                        heap[smallChild] > heap[smallChild+1]:
                    smallChild += 1
                if smallChild >= len(heap) or lastItem <= heap[smallChild]:
                    heap[insertionPoint] = lastItem
                    break
                heap[insertionPoint] = heap[smallChild]
                insertionPoint = smallChild
        return heap[0][1]
	
    def __iter__(self):
        '''Create destructive sorted iterator of priorityDictionary.'''
        def iterfn():
            while len(self) > 0:
                x = self.smallest()
                yield x
                del self[x]
        return iterfn()
	
    def __setitem__(self,key,val):
        '''Change value stored in dictionary and add corresponding
pair to heap.  Rebuilds the heap if the number of deleted items grows
too large, to avoid memory leakage.'''
        dict.__setitem__(self,key,val)
        heap = self.__heap
        if len(heap) > 2 * len(self):
            self.__heap = [(v,k) for k,v in self.iteritems()]
            self.__heap.sort()  # builtin sort likely faster than O(n) heapify
        else:
            newPair = (val,key)
            insertionPoint = len(heap)
            heap.append(None)
            while insertionPoint > 0 and \
                    newPair < heap[(insertionPoint-1)//2]:
                heap[insertionPoint] = heap[(insertionPoint-1)//2]
                insertionPoint = (insertionPoint-1)//2
            heap[insertionPoint] = newPair
	
    def setdefault(self,key,val):
        '''Reimplement setdefault to call our customized __setitem__.'''
        if key not in self:
            self[key] = val
        return self[key]

def Dijkstra(G,start,end=None):
	"""
	Find shortest paths from the start vertex to all
	vertices nearer than or equal to the end.

	The input graph G is assumed to have the following
	representation: A vertex can be any object that can
	be used as an index into a dictionary.  G is a
	dictionary, indexed by vertices.  For any vertex v,
	G[v] is itself a dictionary, indexed by the neighbors
	of v.  For any edge v->w, G[v][w] is the length of
	the edge.  This is related to the representation in
	<http://www.python.org/doc/essays/graphs.html>
	where Guido van Rossum suggests representing graphs
	as dictionaries mapping vertices to lists of neighbors,
	however dictionaries of edges have many advantages
	over lists: they can store extra information (here,
	the lengths), they support fast existence tests,
	and they allow easy modification of the graph by edge
	insertion and removal.  Such modifications are not
	needed here but are important in other graph algorithms.
	Since dictionaries obey iterator protocol, a graph
	represented as described here could be handed without
	modification to an algorithm using Guido's representation.

	Of course, G and G[v] need not be Python dict objects;
	they can be any other object that obeys dict protocol,
	for instance a wrapper in which vertices are URLs
	and a call to G[v] loads the web page and finds its links.
	
	The output is a pair (D,P) where D[v] is the distance
	from start to v and P[v] is the predecessor of v along
	the shortest path from s to v.
	
	Dijkstra's algorithm is only guaranteed to work correctly
	when all edge lengths are positive. This code does not
	verify this property for all edges (only the edges seen
 	before the end vertex is reached), but will correctly
	compute shortest paths even for some graphs with negative
	edges, and will raise an exception if it discovers that
	a negative edge has caused it to make a mistake.
	"""

	D = {}	# dictionary of final distances
	P = {}	# dictionary of predecessors
	Q = priorityDictionary()   # est.dist. of non-final vert.
	Q[start] = 0
	
	for v in Q:
		D[v] = Q[v]
		if v == end: break
		
		for w in G[v]:
			vwLength = D[v] + G[v][w]
			if w in D:
				if vwLength < D[w]:
					raise ValueError, \
  "Dijkstra: found better path to already-final vertex"
			elif w not in Q or vwLength < Q[w]:
				Q[w] = vwLength
				P[w] = v
	
	return (D,P)
			
def shortestPath(G,start,end):
	"""
	Find a single shortest path from the given start vertex
	to the given end vertex.
	The input has the same conventions as Dijkstra().
	The output is a list of the vertices in order along
	the shortest path.
	"""

	D,P = Dijkstra(G,start,end)
	Path = []
	while 1:
		Path.append(end)
		if end == start: break
		end = P[end]
	Path.reverse()
	return Path

# Simple reachability example
SimpleR  = {'x1' :{'x2' :5		             },
		   'x2' :{'x6' :2                    },
		   'x3' :{'x2' :0,  'x7' :6          },
		   'x4' :{'x3' :1                    },
		   'x5' :{'x1' :11, 'x6' :4          },
		   'x6' :{'x1' :5,  'x7' :2          },
		   'x7' :{'x4' :2,  'x8' :3,  'x11':0},
		   'x8' :{'x4' :1,  'x12':4          },
		   'x9' :{'x5' :2,  'x10':3          },
		   'x10':{'x6' :1,  'x7' :2,  'x11':0},
		   'x11':{'x4' :2                    },
		   'x12':{'x11':1                    },
		}

# Two Tanks example
TwoTanks = {'h11':{'h21':1,'h12':1},
			'h21':{'h31':1,'h22':1},
			'h31':{'h41':1,'h32':1},
			'h41':{'h51':1,'h42':1}, 
			'h51':{'h52':1},

			'h12':{'h22':1,'h13':1,'h11':1},
			'h22':{'h32':1,'h23':1,'h21':1},
			'h32':{'h42':1,'h33':1,'h31':1},
			'h42':{'h52':1,'h43':1,'h41':1, 'h32':1}, 
			'h52':{'h53':1,'h51':1,'h42':1},

			'h13':{'h23':1,'h14':1,'h12':1},
			'h23':{'h33':1,'h24':1,'h22':1},
			'h33':{'h43':1,'h34':1,'h32':1},
			'h43':{'h53':1,'h44':1,'h42':1}, 
			'h53':{'h54':1,'h52':1},

			'h14':{'h24':1,'h15':1,'h13':1},
			'h24':{'h34':1,'h25':1,'h23':1},
			'h34':{'h44':1,'h35':1,'h33':1},
			'h44':{'h54':1,'h45':1,'h43':1}, 
			'h54':{'h55':1,'h53':1},

			'h15':{'h25':1,'h14':1},
			'h25':{'h35':1,'h24':1},
			'h35':{'h45':1,'h34':1},
			'h45':{'h55':1,'h44':1}, 
			'h55':{'h54':1},
		
		}

print shortestPath(SimpleR, 'x5', 'x2')
print shortestPath(TwoTanks, 'h55', 'h44')